import sys

import math
import cairo
import logging

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(format=FORMAT)

DEFAULT_CHARSET="cp850"

from .page import LetterPage, A4Page

"""
Thoughts:
create a document suitable for the presenter type
as you stream the print file, add elements to the 
document, presenting as it streams

Think of what a "page" is; this is relevant even 
for interactive presenters, like Terminal

Scaling is also a factor; old documents think in
terms of pages, and they don't necessarily do
things like proper form feeds. I see that with 
AtariWriter+, which seems to walk donw the page
a bit.

"""

class BasePresenter(object):
    """
    Abstract class for presenter objects
    """
    page_list = []
    default_page_size = "A4"    
    cur_page = 0
    dpi = 300
    linespacing = 12
    font_height = 8
    page_lines = None
    superscript = False
    subscript = False
    underline = False


    font_state = {
        "bold": False,
        "italic": False,
        "family": "monospace",
        "size": 10,
    }
    
    print_stream = None
    
    def __init__(self, stream, **kwargs):
        self.print_stream = stream
        

    def initalize(self):
        """
        """
        raise NotImplementedError
    

    def set_underline(self, enable):
        """
        """
        raise NotImplementedError
    
    def carriage_return(self):
        """
        """
        raise NotImplementedError

    def set_condensed(self, enable):
        """
        """
        raise NotImplementedError

    def set_font_size(self, *args):
        """
        """
        raise NotImplementedError


    def set_low_quality(self, enable):
        """
        """
        raise NotImplementedError


    def set_bold(self, value):
        """
        """
        self.bold = value
    
    def newline(self):
        """
        """
        self.carriage_return()
        self.linefeed()

    def set_italic(self, value):
        """
        """
        self.italic = value    
    def set_family(self, enable):
        raise NotImplementedError
    
    def set_font_height(self, value):
        self.font_height = value
    
    def set_font_width(self, value):
        self.font_width = value

    def set_size(self, size):
        """
        """
        raise NotImplementedError
    def set_proportional(self, value):
        """
        """
        self.proportional = value        
    def handle_formfeed(self, count=1):
        """
        """
        raise NotImplementedError
    
    def add_text(self, text):
        """
        """
        raise NotImplementedError

    def set_linespacing(self, linespacing):
        """
        set my linespacing in points
        """
        raise NotImplementedError

    def set_charset(self, charset):
        raise NotImplementedError

    def set_hpos(self, pos):
        raise NotImplementedError
        
    def backspace(self, numspaces=1):
        raise NotImplementedError

class color:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'

        
class PlainTextPresenter(BasePresenter):
    """
    Prints text with no formatting codes, just basic margins
    """
    outfile = None
    linespacing = 1 # character cells, eh
    page_width = 72    
    
    def __init__(self, **kwargs):
        """
        """
        self.outfile = kwargs.get('outfile', sys.stdout)
        size = kwargs.get("size", self.default_page_size)
        self.new_page(size)        
        self.separator = kwargs.get("separator", "\f")

    def new_page(self, size):
        logger.debug("pdf::new_page entered with %s", size)
        new_page = None
        if size == "Letter":
            new_page = LetterPage()
        else:
            new_page = A4Page()

        self.page_list.append(new_page)
        self.cur_page = len(self.page_list) - 1

    def add_text(self, byte):
        if type(byte) == type(b""):
            byte = byte.decode(DEFAULT_CHARSET)
        if byte == "\n":
            byte = byte * self.linespacing
        self.page_list[self.cur_page].text = \
            self.page_list[self.cur_page].text + byte

    def set_bold(self, value):
        """
        """
        if value:
            self.add_text(color.BOLD)
        else:
            self.add_text(color.END)
    def set_condensed(self, value):
        """
        """
        logger.debug("plaintext::set_condensed entered with %s (n/a/)", value)
        return

    def set_underline(self, value):
        """
        """
        if value:
            self.add_text(color.UNDERLINE)
        else:
            self.add_text(color.END)
    
    def __str__(self):
        return "\f".join([t.text for t in self.page_list])
    
    def set_hpos(self, hpos):
        """
        workaround for now
        """
        self.add_text(" ")
    
    def linefeed(self):
        """
        """
        self.add_text("\n")

    def carriage_return(self):
        self.add_text("\r")
    
    def set_linespacing(self, linespacing):
        """
        set my linespacing in points
        """
        self.linespacing = self.linespacing
    
    def set_font_size(self, *args):
        """
        """
        pass

    def set_low_quality(self, value):
        logger.debug("plaintext::set_low_value entered with %s", value)

    def set_charset(self, value):
        """
        """
        logger.debug("plaintext::charset entered with %s", value)
        
class TerminalPresenter(BasePresenter):
    """
    terminal viewer for documents
    """

class PdfPresenter(BasePresenter):
    """
    PDF-Based presenter
    """
    default_font_family = "Courier New"
    authentic = True
    default_font_size = 8
    default_line_height = 8
    default_proportional = True
    condensed = False
    bold = False
    italic = False
    font_width = None    
    page_list = []
    page_size = None
    page_lines = 9999
    cur_line = 0
    cur_page = None
    stretch_x = 1.0
    stretch_y = 1.0
    x = 0
    y = 0
    linespacing = 72.0 / 6.0 # defaults to 1/"
    
    def __init__(self, **kwargs):
        """
        """
        size = kwargs.get("size", self.default_page_size)
        # self.filename = kwargs.get("filename", "default.pdf")
        self.fd = kwargs.get("fdout")
        logger.debug("pdf::init using %s", self.fd)
        logger.debug("pdf::init kwargs %s", kwargs)
        self.font_family = self.default_font_family
        self.font_size = self.default_font_size
        self.proportional = self.default_proportional
        self.page_size = size
        self.em = 0	# width of em-space
        self.en = 0 	# width of en-space

        if size == "Letter":
            self.page = LetterPage()
        else:
            self.page = A4Page()        

        # defaults
        

        # set up cairo stuff
        # self.ps = cairo.PDFSurface(self.fd or self.filename, self.page.width, self.page.height)
        self.ps = cairo.PDFSurface(self.fd, self.page.width, self.page.height)
        self.ctx = cairo.Context(self.ps)
        self.ctx.select_font_face(self.font_family)
        self.ctx.set_font_size(self.font_size)
        self.ctx.set_source_rgb(0, 0, 0)
        self.line_height = self.default_line_height        
        self.home()
        # print(self.x, self.y)

    def set_font_size(self, size):
        self.font_size = size
        

    def newline(self):
        """
        """
        self.carriage_return()
        self.linefeed()

    def carriage_return(self):
        """
        """
        self.x = 0   
        self.cur_line = self.cur_line + 1
        if self.cur_line and self.cur_line > self.page_lines:
            logger.debug("carriage_return out of lines %s/%s", self.cur_line, self.page_lines)
            self.cur_line = 0
            self.new_page()
        
    def linefeed(self):
        """
        """
        logger.debug("pdf::linefeed entered")
        self.y = self.y + self.linespacing * self.stretch_y # + self.line_height * self.stretch_y
        if (self.y + self.page.margin_b) > (self.page.height - self.page.margin_b):
            self.new_page()
        
    def set_hpos(self, pos):
        """
        """
        self.x = pos
    
    def home(self):
        """
        """
        self.y = self.font_size
        self.x = 0

    def add_text(self, byte):
        # do some text handling...
        text = byte.decode(DEFAULT_CHARSET)
        if 0x00 <= ord(text) < 0x1f:
            if text == "\n":
                self.linefeed()
                return
            elif text == "\r":
                self.carriage_return()
                return
            elif text == "\f":
                self.new_page()
                return
            else:
                logger.warn("Not adding text: %s", text)
            logger.debug("Can't print character %s", byte)
            return

        self.set_font()        
        x_advance = self.get_x_advance(text)
        
        if (self.x + x_advance > self.page.width - self.page.margin_r):
            self.newline()
            self.carriage_return()
            self.linefeed()
        scale_modifier = 1.0
        if self.superscript:
            # move up a litte
            y_modifier = self.font_size/-3
        elif self.subscript:
            y_modifier = self.font_size/3 # self.font_size
        else:
            y_modifier = 0

        self.ctx.save()
        self.ctx.move_to(self.page.margin_l + self.x, self.page.margin_t + self.y + self.font_size + y_modifier)
        self.ctx.scale(self.stretch_x, self.stretch_y)
        self.ctx.show_text(text)
        if self.underline:
            # draw a line there...
            start_y = self.y + self.font_size * 1.2 # just an estimate
            start_x = self.x
            self.ctx.move_to(start_x, start_y)
            self.ctx.line_to(start_x + x_advance, start_y)
            self.ctx.set_line_width(0.2)
            self.ctx.stroke()
        self.ctx.restore()
        self.x = self.x + x_advance
        logger.debug("T:'%s' x:%s, y:%s p:%s x_adv: %s b:'%s' l:%s ln:%s/%s lsp:%s x<->:%s y<->:%s sm:%s sub:%s sup:%s" % (text, int(self.x), int(self.y), self.proportional, int(x_advance), self.bold, self.line_height, self.cur_line, self.page_lines, self.linespacing, self.stretch_x, self.stretch_y, scale_modifier, self.superscript, self.subscript))

    def get_x_advance(self, text, proportional=True):
        """
        """
        if proportional and self.proportional:
            res = self.ctx.text_extents(text).x_advance or \
                self.ctx.text_extents(" ").x_advance
        else:
            # update em
            self.em = self.ctx.text_extents("M").width
            res = self.em
        return res * self.stretch_x
        
        
    def set_font(self):
        logger.debug("pdf::set_font entered")
        if self.italic:
            slant =  cairo.FONT_SLANT_ITALIC
        else:
            slant = cairo.FONT_SLANT_NORMAL
                    
        if self.bold:
            weight = cairo.FONT_WEIGHT_BOLD
        else:
            weight = cairo.FONT_SLANT_NORMAL
        
        self.ctx.select_font_face(self.font_family, slant, weight)
        logger.debug("font_size: %s", self.font_size)
        self.ctx.set_font_size(self.font_size)
        self.ctx.set_source_rgb(0, 0, 0)

        # calculate em and en
        self.em = self.get_x_advance("M", proportional=False)
        logger.debug("EM: %d", self.em)
        self.en = self.get_x_advance("N", proportional=False)
        logger.debug("EN: %d", self.em)
            
    def new_page(self):
        """
        """
        logger.debug("pdf::new_page entered")
        self.cur_line = 0
        self.ctx.show_page()
        self.home()

    def set_low_quality(self, *args):
        """
        """
        pass

    def set_underline(self, value):
        self.underline = value

    def set_bold(self, value):
        self.bold = value

    def set_italic(self, value):
        self.italic = value

    def set_proportional(self, value):
        logger.debug("pdf::set_proportional entered with %s", value)
        if value:
            logger.debug("pdf::setting sans-serif")
            self.ctx.select_font_face("sans-serif")
        else:
            logger.debug("pdf::setting %s", self.default_font_family)
            self.ctx.select_font_face(self.default_font_family)

    def set_linespacing(self, value):
        logger.debug("pdf::set_linespacing entered with %s", value)
        # self.linespacing = value - self.font_size + 2
        self.linespacing = value
        self.line_height = self.font_size
    
    def save(self):
        self.ctx.show_page()
    
    def set_charset(self, value):
        logger.debug("chrst: %s", value)

    def set_condensed(self, value):
        logger.debug("pdf::set_condensed entered with %s", value)
        self.stretch_x = value and 0.5 or 1.0
    
    def backspace(self, value):
        """
        moves back x. em-spaces
        """
        logger.debug("pdf::backspace entered")
        self.x = self.x - value * self.em
    
class HtmlPresenter(BasePresenter):
    default_font_family = "monospace"
    body = ""
    
    def __init__(self, **kwargs):
        self.font_family = self.default_font_family

    def linefeed(self):
        self.body = self.body + "</p><p>"

    def set_charset(self, value):
        self.charset = value
        
    def add_text(self, byte):
        text = byte.decode(DEFAULT_CHARSET)
        if text == "\n":
            self.linefeed()
            return
        self.body = self.body + text
    
    def save(self):
        print(self.body)
    
    def set_linespacing(self, value):
        self.linespacing = value
        


import cairo

DEFAULT_MARGIN = 0 * 72 # 1/4" default margin in points

def pts2mm(pts):
    return float(pts) * 0.352778

def pts2in(pts):
    return float(pts) * 0.0138889


class Page(object):
    height = 0
    width = 0
    margin_l = 0
    margin_r = 0
    margin_t = 0
    margin_b = 0
    
    text = ""
    
    def __init__(self, **kwargs):
        self.margin_l = kwargs.get("margin_l", DEFAULT_MARGIN)
        self.margin_r = kwargs.get("margin_r", DEFAULT_MARGIN)
        self.margin_t = kwargs.get("margin_t", DEFAULT_MARGIN)
        self.margin_b = kwargs.get("margin_b", DEFAULT_MARGIN)
    
    @property
    def height_points(self):
        return self.height

    @property
    def height_inches(self):
        return self.height / 72.0
    
    @property
    def height_mm(self):
        return pts2mm(self.height)

    @property
    def height_inches(self):
        return self.width / 72.0
    
    @property
    def height_mm(self):
        return pts2mm(self.width)
    
class LetterPage(Page):
    height = 72 * 11 # points high
    width = 72 * 8.5 # points wide

class A4Page(Page):
    height = 72 * 11.69 # points high
    width = 72 * 8.27 # points wide


     
if __name__=="__main__":
    a4 = A4Page()
    letter = LetterPage()
    
    print("A4 page:")
    print("height: {}mm".format( pts2mm(a4.height)))
    print("width: {}mm".format( pts2mm(a4.width)))
    
    print("Letter page:")
    print("height: {}in".format( pts2in(letter.height)))
    print("width: {}in".format( pts2in(letter.width)))
    print("top margin: {}in".format( pts2in(letter.margin_t)))
    print("bottom margin: {}in".format( pts2in(letter.margin_b)))

    print("Letter page:")
    print("height: {}mm".format( pts2mm(letter.height)))
    print("width: {}mm".format( pts2mm(letter.width)))
    print("top margmm: {}mm".format( pts2in(letter.margin_t)))
    print("bottom margin: {}in".format( pts2in(letter.margin_b)))
    
    


__printer_class__="Atari 8-series"
import sys
import os
import logging

from pyretroprint.presenter import PlainTextPresenter, PdfPresenter
from pyretroprint.page import LetterPage, A4Page


ESCAPE =  b'\x1b'
DEFAULT_CODEPAGE="cp850"

LOGGER=logging.getLogger(__name__)

class AtariProcessor(object):
    """
    atari 8-series processor
    - Atari 825
    """
    escape_state = False
    
    params_count = {
        0x08: 1,
        '-': 1,
    }
    
    def __init__(self, printfile, presenter, size="Letter"):
        """
        """
        self.printfile = printfile
        self.presenter = presenter
        self.presenter.set_linespacing(12) # 1/12" between lines?
        
    def handle_command(self, byte, params):
        """
        """
        command = byte # .decode(DEFAULT_CODEPAGE)
        msg = ""
        if command == b'\x01':
            msg = "Insert one dot space"
            self.send_dot_spaces(1)
            return
            
        elif command == b'\x02':
            msg = "Insert two dot spaces"
            self.send_dot_spaces(2)
            return
        elif command == b'\x03':
            msg = "Insert three dot spaces"
            self.send_dot_spaces(3)
            return
        elif command in [b'\x04', 4]:
            msg = "Insert four dot spaces"
            self.send_dot_spaces(4)
            return
        elif command in [b'\x05', 5]:
            msg = "Insert five dot spaces"
            self.send_dot_spaces(5)
            return
        elif command in [b'\x06', 6]:
            msg = "Insert six dot spaces"
            self.send_dot_spaces(6)
            return
        elif command == b'\x0a':
            msg = "Reverse line feed"
        elif command == b'\x0e':
            msg = "Set elongated print"
            self.set_double_width(1)
            return
        elif command == b'\x0f':
            msg = "Stop elongated print"
            self.set_double_width(0)
            return
        elif command == b'\x11':
            msg = "Set proportional"
            self.set_proportional()
            return
        elif command == b'\x13':
            msg = "Set 10cpi monospace"
            self.set_cpi(10)
            return
        elif command == b'\x14':
            msg = "Select condensed characters"
            self.set_condensed(1)
            return
        elif command == b'\x1c':
            msg = "Advance half a line"
        elif command == b'\x1e':
            msg = "Reverse half a line"
        elif command == b'-':
            msg = "Handle underline"
            self.set_underline(params[0])
            return
        elif command == b'\x34':
            msg = "Start italics"
            self.set_italic(1)
            return
        elif command == b'\x35':
            msg = "Stop italics"
            self.set_italic(0)
            return
        else:
            msg = "Unexpected command: %s" % command
        LOGGER.debug("Not handled: ESC %s %s %s", byte, params, msg)

    def set_linespacing(self, value):
        """
        """
        LOGGER.debug("atari::set_linespacing entered with %s", value)
        self.presenter.set_linespacing(value) # 1/12" b/w lines?

    def set_proportional(self):
        """
        """
        LOGGER.debug("atari::set_proportional entered")
        self.presenter.set_proportional(1)

    def set_italic(self, value):
        LOGGER.debug("atari::set_italic entered with %s", value)
        self.presenter.set_italic(value)

    def send_dot_spaces(self, value):
        """
        """
        LOGGER.debug("atari::send_dot_spaces entered with %s", value)
        self.presenter.x = self.presenter.x + value * 0.95 * self.presenter.stretch_x # needs to be fine tuned a little bit
    def set_bold(self, value):
        """
        """
        LOGGER.debug("atari::set_bold entered with %s", value)
        self.presenter.set_bold(value)
    def set_condensed(self, value):
        """
        """
        LOGGER.debug("atari::set_condensed entered with %s", value)
        self.presenter.set_condensed(value)

    def set_cpi(self, value):
        """
        there is only 10 or 12 here...
        """
        LOGGER.debug("atari::set_cpi entered with %s", value)
        if value == 10:
            LOGGER.debug("atari::set_cpi setting 12 pt")
            self.presenter.set_font_size(12) # 12 points in courier is about 10 cpi
        elif value == 12:
            LOGGER.debug("atari::set_cpi setting 10 pt")
            self.presenter.set_font_size(10)
        else:
            LOGGER.warn("atari::set_cpi called with unhandled value %s", value)
            sys.exit(1)
        self.presenter.stretch_x = 1.0        
        self.presenter.set_proportional(0)
            
    def set_double_width(self, value):
        LOGGER.debug("atari::set_double_width entered with %s", value)
        if value:
            self.presenter.stretch_x = 2.0
        else:
            self.presenter.stretch_x = 1.0


    def handle_escape(self):
        """
        what to do when escape is sent
        """
        LOGGER.debug("atari::handle_escape entered")
        self.escape_state = True
        
        
    def process(self):
        """
        main processing loop
        """
        LOGGER.debug("atari::process entered")
        pos = 0
        while True:
            byte = self.printfile.read(1)
            pos = pos + 1
            LOGGER.debug("pos:%s:%s", pos, byte)
            if self.escape_state:
                params = []
                count = self.params_count.get(byte.decode(DEFAULT_CODEPAGE), 0)
                while count:
                    params.append(int.from_bytes(self.printfile.read(1), "big"))
                    count = count - 1
                
                self.escape_state = False
                self.handle_command(byte, params)
            elif byte == ESCAPE:
                self.handle_escape()
            elif byte == b'\x08':
                LOGGER.debug("Backspace n spaces")
                num_spaces = int.from_bytes(self.printfile.read(1), "big")
                self.send_dot_spaces(-num_spaces)
            elif byte == b'\x0a':
                LOGGER.debug("Advance one line")
                self.presenter.linefeed()
            elif byte == b'\x0d':
                self.presenter.carriage_return()
            elif byte == b'\x0e':
                LOGGER.debug("Stop underlining")
                self.set_underline(0)
            elif byte == b'\x0f':
                LOGGER.debug("Start underlining")
                self.set_underline(1)
            elif byte == b'\x9b':
                LOGGER.debug("nbsp/lf")
                self.presenter.newline()
            elif byte == '' or len(byte)==0:
                LOGGER.debug("EOF")
                break
            else:
                self.handle_byte(byte)

    def handle_byte(self, byte):
        self.presenter.add_text(byte)

    def set_underline(self, value):
        LOGGER.debug("atari::set_underline entered with %s", value)
        if value:
            self.presenter.set_underline(True)
        else:
            self.presenter.set_underline(False)

    def backspace(self, num_spaces):
        LOGGER.debug("atari::backspace entered with %s", num_spaces)
        self.presenter.backspace(num_spaces)

if __name__=="__main__":
    # getopt etc.

    printfile = open(sys.argv[-1], 'rb')
    print("Reading", sys.argv[-1])
    print("Processing for PDF")
    presenter = PdfPresenter()
    proc = AtariProcessor(printfile, presenter, size="Letter")
    proc.process()
    print(proc.presenter)
    


__printer_class__="Epson FX-80"
import sys
import os
import logging

from pyretroprint.presenter import PlainTextPresenter, PdfPresenter, HtmlPresenter
import pyretroprint.page

ESCAPE = 0x1b

CODES = {
    
}

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
FORMAT = '%(asctime)-15s %(message)s'
logging.basicConfig(format=FORMAT)


class PrintFile(object):
    """
    """
    def __init__self(fd):
        """
        """
        self.fd = fd

    def read(self, bytes=1):
        next_byte = self.fd.read(bytes)

class EpsonProcessor(object):
    """
    esc/p and esc/p 2 processor
    """
    clear_on_line = []
    escape_state = None
    commands = {
        "%": {
                "params_count": 1,
                "description": "Select user-defined set"
            },
        "$": {
                "params_count": 2,
                "description": "ESC $: set abs. HPOS"
            },
        "t": 1,
        "R": 1,
        "k": 1,
        "!": 1,
        "x": 1,
        "C": 1        
        
    }
    params_count = {
        "%": 1,
        "$": 2,
        "t": 1,
        "R": 1,
        "k": 1,
        "!": 1,
        "x": 1,
        "C": 1,
        "+": 1,
        "p": 1,
        "3": 1,
        "-": 1,
        "S": 1,
        "L": "NLDK",
        "Y": "NLDK",
        "Z": "NLDK",
        "D": None, # NUL terminated
        
    }
    
    params = []
    
    defined_unit = 72.0 / 60.0 # 1 / 60 in. in points
    
    def __init__(self, pf, presenter, size="Letter"):
        """
        """
        self.printfile = pf
        self.presenter = presenter # initialized presenter
        self.initialize_printer()

    def handle_command(self, byte, params):
        """
        """
        command = byte.decode("cp850")
        msg = ""
        if command == "%":
            msg = "Select user-defined set"
        elif command == "@":
            msg = "Initialize printer"
            self.initialize_printer()
            return
        elif command == "$":
            msg = "Set abs. HPOS"
            self.set_hpos(params)
            return
        elif command == "t":
            msg = "Select character table"
        elif command =="6":
            msg = "Enable printing of upper control codes"
            self.enable_upper()
            return
        elif command =="R":
            msg = "Select an international character set"
            self.select_intl_charset(params[0])
            return
        elif command =="2":
            msg = "Select 1/6-inch line spacing"
            self.set_lpi(6)
            return
        elif command == "k":
            msg = "Select typeface"
        elif command == "!":
            msg = "Master select"
            self.master_select(params[0])
            return
        elif command == "x":
            msg = "Select LQ or Draft"
            self.set_lq(params[0])
            return
        elif command == "C":
            msg = "Set page length in lines"
            self.set_page_lines(params[0])
            return
        elif command == "+":
            msg = "Set n/360th in. line spacing"
            self.set_linespacing(params[0], 360)
            return
        elif command == "E":
            msg = "Set the weight attribute of the font to bold"
            self.set_bold(1)
            return
        elif command == "F":
            msg = "Cancel bold font"
            self.set_bold(0)
            return
        elif command == "p":
            msg = "Turn proportional off/on"
            self.set_proportional(params[0])
            return
        elif command == "T":
            msg = "Cancel super-/subscript"
            self.set_supersub(-1)
            return
            
        elif command == "P":
            msg = "Select 10.5-point, 10-cpi"
            self.set105_10cpi()
            return
        elif command == "3":
            msg = "Set n/216-inch line spacing"
            self.set_linespacing(params[0], 216)
            return
        elif command == "Y" or command == "L":
            msg = "Set 120 dpi double-speed graphics"
            self.set_dpi(120, params)
            return
        elif command == "4":
            msg = "Set italic"
            self.set_italic(1)
            return
        elif command == "G":
            msg = "Set doublestrike"
            self.set_bold(1)
            return
        elif command == "H":
            msg = "Cancel doublestrike"
            self.set_bold(0)
            return
    
        elif command == "Z":
            msg = "Set 240 dpi graphics %s bytes" % (len(params))
            self.set_dpi(240, params)
            return
        elif command == "5":
            msg = "Cancel italic"
            self.set_italic(0)
            return
        elif command == "-":
            msg = "Cancel italic"
            self.set_underline(params[0])
            return
        elif command == "D":
            msg = "Set tabs"
            self.set_tabs(params)
            return
        elif command == "S":
            msg = "Select superscript/subscript printing"
            self.set_supersub(params[0])
            return
        else:
            msg = "ESC %s not handled" % command
        logger.debug("Not handled: ESC %s %s %s", command, params, msg)

    def set_page_lines(self, value):
        """
        """
        self.page_lines = value
        self.presenter.page_lines = value

    def handle_escape(self):
        """
        """
        self.escape_state = True

    def set_tabs(self, params):
        """
        """
        logger.debug("epson::set_tabs entered")

    def set_supersub(self, value):
        """
        """
        logger.debug("epson::set_supersub entered with %s", value)
        if value in [1, 49]:
            logger.debug("epson::set_supersub setting super")
            self.presenter.superscript = True
            self.presenter.subscript = False
            return
        if value in [0, 48]:
            logger.debug("epson::set_supersub setting sub")
            self.presenter.subscript = True 
            self.presenter.superscript = False 
            return
        
        self.presenter.superscript = False            
        self.presenter.subscript = False

    def master_select(self, value):
        """
        """
        logger.debug("epson::master_select entered")
        value & 1 and  \
            self.presenter.set_font_width(7.2)  \
            or self.presenter.set_font_width(6) 
        
        self.set_proportional(value & 2)
        self.set_condensed(value & 4)
        self.set_bold(value & 8)
        self.set_bold(value & 16)
        self.set_double_width(value & 32)
        self.set_italic(value & 64)
        self.set_underline(value & 128)

    def set_lq(self, value):
        """
        """
        logger.debug("epson::set_lq entered")
        value in [0, 48] and self.presenter.set_low_quality(1)
        value in [1, 49] and self.presenter.set_low_quality(0)

    def set_condensed(self, value):
        """
        """
        logger.debug("epson::set_condensed entered with %s", value)
        self.presenter.set_condensed(value)
        
    def set105_10cpi(self):
        """
        
        """
        logger.debug("epson::set105_10cpi entered")
        # 72 points per inch, 10 CPI would be 7.2pts right?
        self.presenter.set_font_size(10.5)
        self.presenter.stretch_x = 1.0
        self.presenter.set_font_width(7.2)        
        

    def set_double_width(self, value):
        logger.debug("epson::set_double_width entered")
        if value:
            self.presenter.stretch_x = 2.0
        else:
            self.presenter.stretch_x = 1.0
    
    def set_underline(self, value):
        logger.debug("epson::set_underline entered")
        if value in [1,49]:
            self.presenter.set_underline(True)
        else:
            self.presenter.set_underline(False)
    
    def select_intl_charset(self, charset):
        """
        0 USA
        1 France
        2 Germany
        3 United Kingdom
        4 Denmark I
        5 Sweden
        6 Italy
        7 Spain I
        8 Japan (English)
        9 Norway
        10 Denmark II
        11 Spain II
        12 Latin America
        13 Korea
        64 Legal
        """
        logger.debug("epson::set_intl_charset entered")
        self.presenter.set_charset(charset)
        
    def initialize_printer(self):
        """
        """
        logger.debug("epson::initialize_printer entered")
        self.set_bold(0)
        self.set_proportional(0)
        self.set_italic(0)
        self.defined_unit = 72.0 / 60.0
        self.presenter.set_font_size(10.5)
        self.set_linespacing(1, 6)

    def enable_upper(self):
        """
        """
        logger.debug("epson::enable_upper entered")
        
    def set_proportional(self, value):
        """
        """
        logger.debug("epson::set_proportional entered w/ %s", value)
        prop = int(value) in [1,49]
        self.presenter.set_proportional(prop)

    def set_italic(self, value):
        """
        """
        logger.debug("epson::set_italic entered")
        self.presenter.set_italic(value)
    
    def set_lpi(self, value):
        """
        """
        logger.debug("epson::set_lpi entered with %s", value)
        linespacing = 72.0/value  # value in points
        self.presenter.set_linespacing(linespacing)

    def set_dpi(self, value, params):
        """
        """
        # not really relevant on a virtual printer...
        logger.debug("epson::set_dpi entered with %s", value)
        self.presenter.dpi = value
        logger.debug("%s bytes of data for %s", len(params), value)
        if len(params):
            logger.debug("%s...%s", params[0], params[-1])

    def set_bold(self, value):
        """
        """
        logger.debug("epson::set_bold entered")
        self.presenter.set_bold(value)
    
    def set_hpos(self, params):
        """
        """
        logger.debug("epson::set_hpos entered")
        
        nL = params[0]
        nH = params[1]
        
        lm = self.presenter.page.margin_l
        # lm = self.presenter.page_list[self.presenter.cur_page].margin_l
        du = self.defined_unit
        
        hpos = (((nH * 256) + nL) * du) + lm
        logger.debug("Setting hpos to %s", hpos)
        self.presenter.set_hpos(hpos)

    def set_linespacing(self, param, base):
        """
        Sets the line spacing in points on the presenter
        """
        logger.debug("epson::set_linespacing entered: %s/%s", param, base)
        n = param
        sp_inches = n / base
        logger.debug("sp_inches = %s", sp_inches)
        # self.set_lpi(sp_inches)
        self.presenter.set_linespacing(sp_inches * 72.0) # linespacing in points

    def process(self):
        """
        """
        logger.debug("epson::process entered")
        logger.debug("Processing...")
        while True:
            byte = self.printfile.read(1)
            if self.escape_state:
                params = []
                count = self.params_count.get(byte.decode("cp850"), 0)
                
                while count and count not in  ["NLDK",]:
                    params.append(int.from_bytes(self.printfile.read(1), "big"))
                    count = count - 1
                    
                if count == None:
                    # null termed parm list
                    # read params until it's NUL
                    parm = None
                    while True:
                        parm = self.printfile.read(1)
                        if parm == b'\x00':
                            break
                        params.append(parm)
                elif count == "NLDK":
                    logger.debug("NLDK %s", byte)
                    
                    params = []
                    nL = int.from_bytes(self.printfile.read(1), "big") # 
                    logger.debug("nL:%s", nL)
                    nH = int.from_bytes(self.printfile.read(1), "big") # 
                    logger.debug("nH:%s", nH)
                    graphcols = nH * 256 + nL

                    i = 0
                    quick = False # ESC Y skips every other dot (urgh)
                    this_byte = None
                    if byte in [b'Y']:
                        quick = True
                    logger.debug("q:%s graphcols:%s", quick, graphcols)
                    while i < int(graphcols):
                        # get d1..dk
                        this_byte = self.printfile.read(1)
                        params.append(this_byte)
                        i = i + 1
                    logger.debug("Got %d bytes graphics data", len(params))
                    
                self.escape_state = False
                self.handle_command(byte, params)
            elif byte == b'\x1b':
                self.handle_escape()
            elif byte == b'\x0e':
                self.set_double_width(True)
                self.clear_on_line.append(self.set_double_width)
            elif self.escape_state:
                self.escape_state = False
                self.handle_command(byte)
            elif byte == b'':
                break
            # check for some control codes
            elif byte == b'\x9b':
                self.presenter.newline()
                [f(False) for f in self.clear_on_line]
                self.clear_on_line = []
            elif byte == b'\x0f':
                logger.debug("XXX: SETTING CONDENSED")
                self.set_condensed(True)
            elif byte == b'\x12':
                self.set_condensed(False)
            else:
                self.handle_byte(byte)
 
    def handle_byte(self, byte):
        self.presenter.add_text(byte)


if __name__=="__main__":
    # getopt etc.
    printfile = open(sys.argv[-1], 'rb')
    print("Reading", sys.argv[-1])
    print("Processing for PDF")
    presenter = PdfPresenter()
    proc = EpsonProcessor(printfile, presenter, size="Letter")
    # proc = EpsonProcessor(printfile, PlainTextPresenter)
    proc.process()
    print(proc.presenter)


"""
Main file for pyretroprint
"""
import sys
import argparse
import io
import logging

from pyretroprint.epsonfx import EpsonProcessor
# from pyretroprint.ibm import ProprinterProcessor
from pyretroprint.atari import AtariProcessor
from pyretroprint.presenter import PdfPresenter
from pyretroprint.presenter import PlainTextPresenter

LOGGER = logging.getLogger(__name__)

class PrintOutFile:
    """
    Context manager for handling output files
    """
    filename = None
    file = None
    def __init__(self, printfile):
        """
        setup
        """
        # if this is buffered writer, return it
        LOGGER.debug("printoutfile: __init__ called (%s)", printfile)
        if isinstance(printfile, io.BufferedWriter) or printfile is None:
            LOGGER.debug("printoutfile: Detected io.BufferedWriter")
            self.file = sys.stdout.buffer
        # if it's a pathlike, return binary handle
        elif isinstance(printfile, str):
            LOGGER.debug("printoutfile: Detected filename")
            self.filename = printfile
        else:
            LOGGER.debug("printoutfile: Couldn't detect filetype")
            self.file = sys.stdout.buffer

    def __enter__(self):
        """
        called when context manager started
        """
        LOGGER.debug("printoutfile: __enter__ called (%s)", self.file)
        if self.filename:
            self.file = open(self.filename, 'wb')
        return self.file

    def __exit__(self, exc_type, exc_val, exc_tb):
        LOGGER.debug("printoutfile: exc_type %s", exc_type)

def main():
    """
    Main program logic
    """
    parser = argparse.ArgumentParser(
        description='Process old-school printer jobs'
    )
    parser.add_argument("-v", "--verbose", help="increase output verbosity",
                        action="store_true")
    parser.add_argument("--presenter", "-p",
                        help="name of target output, plaintext or pdf",
                        default="terminal")
    parser.add_argument("--size", "-s", help="Page size for output, a4 or letter",
                        default="Letter")
    parser.add_argument("--printer", "-P",
                        help="Default printer type to emulate: epson, atari, or ibm",
                        default="epson")
    parser.add_argument("--output", "-o", help="Output file, defaults to stdout")
    parser.add_argument("--input", "-i", help="Input file, defaults to stdin")

    args = parser.parse_args()

    processors = {
        "epson": EpsonProcessor,
        # "ibm": IbmProcessor,
        "atari": AtariProcessor
    }

    proc = processors.get(args.printer)
    if not proc:
        sys.stderr.write(f"Unknown printer: {args.printer}")
        sys.stderr.flush()
        sys.exit(1)

    fdoutput = sys.stdout.buffer

    if args.output:
        if args.output == "-":
            pass
        else:
            fdoutput = args.output
    LOGGER.debug("main::Using output %s", fdoutput)
    if args.input:
        fdinput = args.input
    else:
        fdinput = 0 # stdin

    if args.presenter.lower()=="pdf":
        pres = PdfPresenter
    else:
        pres = PlainTextPresenter

    with open(fdinput, 'rb') as infile, PrintOutFile(args.output) as outfile:
        processor = proc(infile, pres(size=args.size, fdout=outfile))
        processor.process()

if __name__=="__main__":
    main()



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

setup(
    name='pyretroprint', 
    version='0.1.7',
    description='PyRetroPrint: Dot-Matrix Emulation for our Bleak Dystopian Future!',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/rubeon/pyretroprint',
    author='rubeon',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    
    keywords='retro, glue, atari, emulation',
    package_dir={'pyretroprint': 'pyretroprint'},
    packages=find_packages(where='.'),
    python_requires='>=3.6, <4',
    install_requires=['pycairo'],
    entry_points={  # Optional
        'console_scripts': [
            'pyretroprint=pyretroprint.pyretroprint:main',
        ],
    },
    
    project_urls={ 
        'Bug Reports': 'https://github.com/rubeon/pyretroprint/issues',
        'Source': 'https://github.com/rubeon/pyretroprint/',
    }
)